/*-
 * Copyright (c) 2015 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Rozhuk Ivan <rozhuk.im@gmail.com>
 *
 */


#include <sys/param.h>
#ifndef BSD
#define _GNU_SOURCE /* See feature_test_macros(7) */
#define __USE_GNU 1
#endif
#include <sys/types.h>

#include <stdlib.h> /* malloc, exit */
#include <unistd.h> /* close, write, sysconf */
#include <fcntl.h> // open
#include <string.h> /* bcopy, bzero, memcpy, memmove, memset, strerror... */
#include <stdio.h> /* for snprintf, fprintf */
#include <time.h>
#include <errno.h>
#include <dirent.h> // opendir, readdir
#include <getopt.h>

#include "mem_find.h"
#include "buf_case.h"
#include "xml.h"


/* From linux: dirent.h */
#ifndef _D_EXACT_NAMLEN
#ifdef _DIRENT_HAVE_D_NAMLEN
# define _D_EXACT_NAMLEN(d) ((d)->d_namlen)
# define _D_ALLOC_NAMLEN(d) (_D_EXACT_NAMLEN (d) + 1)
#else
# define _D_EXACT_NAMLEN(d) (sstrlen ((d)->d_name))
# ifdef _DIRENT_HAVE_D_RECLEN
#  define _D_ALLOC_NAMLEN(d) (((char*) (d) + (d)->d_reclen) - &(d)->d_name[0])
# else
#  define _D_ALLOC_NAMLEN(d)						\
	((sizeof(d)->d_name > 1) ? sizeof (d)->d_name : _D_EXACT_NAMLEN (d) + 1)
# endif
#endif
#endif /* _D_EXACT_NAMLEN */


#ifndef min
#define min(a,b)	(((a) < (b)) ? (a) : (b))
#endif
#define sstrlen(str)	((NULL == (str)) ? 0 : strlen((str)))


#define LOG_ERR(error, descr)						\
	    if (0 != error)						\
		fprintf(stderr, "%s , line: %i, error: %i - %s - %s\n", __FUNCTION__, __LINE__, (error), strerror((error)), (descr))
#define LOG_EV(descr)							\
	    fprintf(stdout, "%s , line: %i: %s\n", __FUNCTION__, __LINE__, (descr))
#define LOG_INFO(descr)							\
	    fprintf(stdout, "%s\n", (descr))
#define LOG_ERR_FMT(error, fmt, args...)				\
	    if (0 != error)						\
		fprintf(stderr, "%s , line: %i, error: %i - %s" fmt "\n", __FUNCTION__, __LINE__, (error), strerror((error)), ##args)
#define LOG_EV_FMT(fmt, args...)					\
	    fprintf(stdout, "%s , line: %i: " fmt "\n", __FUNCTION__, __LINE__, ##args)
#define LOG_INFO_FMT(fmt, args...)					\
	    fprintf(stdout, fmt"\n", ##args)


#define MAX_ARGS_ELEMS		16
#define MAX_MSD_SRCS		16
#define MAX_CFG_FILES		32

typedef struct astra_cfg_channel_s {
	uint8_t		*name;
	size_t		name_size;
	uint8_t		*input;
	size_t		input_size;
	uint8_t		*output;
	size_t		output_size;
	size_t		output_udp_addr_offset;
	int		enable;
} as_cfg_ch_t, *as_cfg_ch_p;

typedef struct astra_cfg_file_s {
	as_cfg_ch_p	channels;
	size_t		channels_cnt;
	void		*buf;
} as_cfg_file_t, *as_cfg_file_p;



typedef struct msd_cfg_source_s {
	uint8_t		*type;
	size_t		type_size;
	uint8_t		*profile_name;
	size_t		profile_name_size;
	uint8_t		*addr;
	size_t		addr_size;
	int		addr_udp;
} msd_cfg_src_t, *msd_cfg_src_p;

typedef struct msd_cfg_channel_s {
	uint8_t		*name;
	size_t		name_size;
	uint8_t		*profile_name;
	size_t		profile_name_size;
	msd_cfg_src_t	srcs[MAX_MSD_SRCS];
	size_t		srcs_cnt;
} msd_cfg_ch_t, *msd_cfg_ch_p;

typedef struct msd_cfg_file_s {
	msd_cfg_ch_p	channels;
	size_t		channels_cnt;
	void		*buf;
	size_t		buf_size;
} msd_cfg_file_t, *msd_cfg_file_p;



typedef struct m3u_channel_s {
	uint8_t		*name;
	size_t		name_size;
	uint8_t		*url;
	size_t		url_size;
	size_t		url_udp_addr_offset;
} m3u_ch_t, *m3u_ch_p;

typedef struct m3u_file_s {
	m3u_ch_p	channels;
	size_t		channels_cnt;
	void		*buf;
} m3u_file_t, *m3u_file_p;


as_cfg_ch_p as_cfg_find_by_addr(as_cfg_file_p as_cfg_files, size_t as_cfg_files_cnt,
	    uint8_t *addr, size_t addr_size);
int	as_cfg_load_dir(char *dir_name, as_cfg_file_p as_cfg_files,
	    size_t *as_cfg_files_cnt);
int	as_cfg_load_file(char *file_name, as_cfg_file_p cfg_file);
int	as_cfg_file_parse(int fd, as_cfg_file_p cfg_file);

msd_cfg_ch_p msd_channel_find_by_addr(msd_cfg_file_p msd_cfg_files, size_t msd_cfg_files_cnt,
	    uint8_t *addr, size_t addr_size);
int	msd_cfg_load_file(char *file_name, msd_cfg_file_p msd_cfg_files,
	    size_t *msd_cfg_files_cnt);

m3u_ch_p pl_m3u_find_by_addr(m3u_file_p m3u_files, size_t m3u_files_cnt,
	    uint8_t *addr, size_t addr_size);
int	pl_m3u_buf_load(char *file_name, m3u_file_p m3u_file);


static int out_fd = -1;




int
main(int argc, char *argv[]) {
	int c, option_index, out_append = 0, name_resolv_pref_astra = 0, verbose = 0;
	size_t i, j, k, found, cnt, tm;
	char *cmd;
	size_t cmd_size;
	char *astra_cfg_dirs_names[MAX_ARGS_ELEMS], *astra_cfg_files_names[MAX_ARGS_ELEMS];
	size_t astra_cfg_dirs_names_cnt = 0, astra_cfg_files_names_cnt = 0;
	char *msd_cfg_files_names[MAX_ARGS_ELEMS];
	size_t msd_cfg_files_names_cnt = 0;
	char *m3u_files_names[MAX_ARGS_ELEMS];
	size_t m3u_files_names_cnt = 0;
	char *grep_arg = NULL, *http_proxy = NULL, *http_proxy_args = NULL;
	size_t grep_arg_size = 0, http_proxy_size = 0, http_proxy_args_size = 0;
	char *out_fn = "/dev/stdout";
	as_cfg_file_t as_cfg_files[MAX_CFG_FILES];
	size_t as_cfg_files_cnt = 0;
	as_cfg_ch_p as_cfg_ch;
	msd_cfg_file_t msd_cfg_files[MAX_CFG_FILES];
	size_t msd_cfg_files_cnt = 0;
	msd_cfg_ch_p msd_cfg_ch;
	m3u_file_t m3u_files[MAX_CFG_FILES];
	size_t m3u_files_cnt = 0;
	m3u_ch_p m3u_ch;
	char name[4096], name2[4096], buf[8192];

	for (;;) {
		static struct option long_options[] = {
			{ "cmd",		required_argument,	0,  0 },
			{ "astra_cfg_dir",	required_argument,	0,  0 },
			{ "astra_cfg_file",	required_argument,	0,  0 },
			{ "msd_cfg_file",	required_argument,	0,  0 },
			{ "m3u_file",		required_argument,	0,  0 },
			{ "grep_arg",		required_argument,	0,  0 },
			{ "http_proxy",		required_argument,	0,  0 },
			{ "http_proxy_args",	required_argument,	0,  0 },
			{ "name_resolv_pref_astra", no_argument,	0,  0 },
			{ "out",		required_argument,	0,  0 },
			{ "out_append",		no_argument,		0,  0 },
			{ "verbose",		no_argument,		0,  0 },
			{ NULL,        		0,			0,  0 }
		};

		option_index = 0;
		c = getopt_long(argc, argv, "", long_options, &option_index);
		if (-1 == c)
			break;
		
		switch (option_index) {
		case 0:
			cmd = optarg;
			cmd_size = sstrlen(cmd);
			break;
		case 1:
			if (MAX_ARGS_ELEMS <= astra_cfg_dirs_names_cnt)
				break;
			astra_cfg_dirs_names[astra_cfg_dirs_names_cnt] = optarg;
			astra_cfg_dirs_names_cnt ++;
			break;
		case 2:
			if (MAX_ARGS_ELEMS <= astra_cfg_files_names_cnt)
				break;
			astra_cfg_files_names[astra_cfg_files_names_cnt] = optarg;
			astra_cfg_files_names_cnt ++;
			break;
		case 3:
			if (MAX_ARGS_ELEMS <= msd_cfg_files_names_cnt)
				break;
			msd_cfg_files_names[msd_cfg_files_names_cnt] = optarg;
			msd_cfg_files_names_cnt ++;
			break;
		case 4:
			if (MAX_ARGS_ELEMS <= m3u_files_names_cnt)
				break;
			m3u_files_names[m3u_files_names_cnt] = optarg;
			m3u_files_names_cnt ++;
			break;
		case 5:
			grep_arg = optarg;
			grep_arg_size = sstrlen(grep_arg);
			break;
		case 6:
			http_proxy = optarg;
			http_proxy_size = sstrlen(http_proxy);
			break;
		case 7:
			http_proxy_args = optarg;
			http_proxy_args_size = sstrlen(http_proxy_args);
			break;
		case 8:
			name_resolv_pref_astra = 1;
			break;
		case 9:
			out_fn = optarg;
			break;
		case 10:
			out_append = 1;
			break;
		case 11:
			verbose = 1;
			break;
		}
	}
	/* open out file. */
	if (0 == out_append) {
		out_fd = open(out_fn, (O_WRONLY | O_CREAT | O_TRUNC), 0600);
	} else {
		out_fd = open(out_fn, (O_WRONLY | O_APPEND), 0600);
		lseek(out_fd, 0, SEEK_END);
	}
	if (-1 == out_fd) {
		LOG_ERR_FMT(errno, "open(%s)", out_fn);
		return (errno);
	}

	/* load astra configs. */
	for (i = 0; i < astra_cfg_dirs_names_cnt; i ++) {
		as_cfg_load_dir(astra_cfg_dirs_names[i], as_cfg_files, &as_cfg_files_cnt);
	}
	for (i = 0; i < astra_cfg_files_names_cnt && MAX_CFG_FILES > as_cfg_files_cnt; i ++) {
		if (0 == as_cfg_load_file(astra_cfg_files_names[i], &as_cfg_files[as_cfg_files_cnt]))
			as_cfg_files_cnt ++;
	}

	/* load astra configs. */
	for (i = 0; i < msd_cfg_files_names_cnt && MAX_CFG_FILES > msd_cfg_files_cnt; i ++) {
		msd_cfg_load_file(msd_cfg_files_names[i], msd_cfg_files, &msd_cfg_files_cnt);
	}

	/* load m3u playlists. */
	for (i = 0; i < m3u_files_names_cnt && MAX_CFG_FILES > m3u_files_cnt; i ++) {
		if (0 == pl_m3u_buf_load(m3u_files_names[i], &m3u_files[m3u_files_cnt]))
			m3u_files_cnt ++;
	}


	/* execute command. */
	if (0 == buf_cmp(cmd, cmd_size, "astra_m3u", 9)) { /* generate m3u */
		if (0 == out_append)
			write(out_fd, "#EXTM3U \r\n\r\n", 12);
		for (i = 0; i < as_cfg_files_cnt; i ++) {
			for (j = 0; j < as_cfg_files[i].channels_cnt; j ++) {
				as_cfg_ch = &as_cfg_files[i].channels[j];
				if (0 == as_cfg_ch->enable)
					continue;
				if (0 != as_cfg_ch->output_udp_addr_offset) {
					uint8_t *ptm = (as_cfg_ch->output + as_cfg_ch->output_udp_addr_offset);
					if (0 == http_proxy_size) {
						if ((sizeof(name) - 16) < as_cfg_ch->output_size)
							continue;
						memcpy(name, "udp://@", 7);
						tm = (as_cfg_ch->output_size - (ptm - as_cfg_ch->output));
						memcpy((name + 7), ptm, tm);
						name[(7 + tm)] = 0;
					} else {
						if ((sizeof(name) - 4) < (http_proxy_size +
						    as_cfg_ch->output_size + http_proxy_args_size))
							continue;
						memcpy(name, http_proxy, http_proxy_size);
						tm = (as_cfg_ch->output_size - (ptm - as_cfg_ch->output));
						memcpy((name + http_proxy_size), ptm, tm);
						if (http_proxy_args_size) {
							name[(http_proxy_size + tm)] = '?';
							memcpy((name + http_proxy_size + tm + 1),
							    http_proxy_args, http_proxy_args_size);
							http_proxy_args_size ++;
						}
						name[(http_proxy_size + tm + http_proxy_args_size)] = 0;
					}
					out_fn = name;
				} else {
					out_fn = (char*)as_cfg_ch->output;
				}
				
				tm = snprintf(buf, sizeof(buf), "#EXTINF:-1,%s\r\n%s\r\n\r\n",
				    as_cfg_ch->name, out_fn);
				write(out_fd, buf, tm);
			}
			write(out_fd, "\r\n\r\n", 4);
		}
	} else if (0 == buf_cmp(cmd, cmd_size, "msd_m3u", 7)) {
		if (0 == out_append)
			write(out_fd, "#EXTM3U \r\n\r\n", 12);
		for (i = 0; i < msd_cfg_files_cnt; i ++) {
			for (j = 0; j < msd_cfg_files[i].channels_cnt; j ++) {
				msd_cfg_ch = &msd_cfg_files[i].channels[j];
				if ((sizeof(name2) - 1) < msd_cfg_ch->name_size)
					continue;
				memcpy(name2, msd_cfg_ch->name, msd_cfg_ch->name_size);
				name2[msd_cfg_ch->name_size] = 0;

				tm = 0;
				found = 0;
				if (name_resolv_pref_astra) {
msd_m3u_resolv_astra:
					tm ++;
					for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt; k ++) {
						as_cfg_ch = as_cfg_find_by_addr(as_cfg_files,
						    as_cfg_files_cnt,
						    msd_cfg_files[i].channels[j].srcs[k].addr,
						    msd_cfg_files[i].channels[j].srcs[k].addr_size);
						if (NULL != as_cfg_ch) {
							found = 1;
							tm = min((sizeof(name) - 1), as_cfg_ch->name_size);
							memcpy(name, as_cfg_ch->name, tm);
							name[tm] = 0;
							break;
						}
					}
				}
				if (0 == found) {
					for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt; k ++) {
						m3u_ch = pl_m3u_find_by_addr(m3u_files, m3u_files_cnt,
						    msd_cfg_files[i].channels[j].srcs[k].addr,
						    msd_cfg_files[i].channels[j].srcs[k].addr_size);
						if (NULL != m3u_ch) {
							found = 1;
							tm = min((sizeof(name) - 1), m3u_ch->name_size);
							memcpy(name, m3u_ch->name, tm);
							name[tm] = 0;
							break;
						}
					}
				}
				if (0 == found && 0 == name_resolv_pref_astra && 0 == tm)
					goto msd_m3u_resolv_astra;
				if (0 == found) {
					found = 1;
					tm = min((sizeof(name) - 1), msd_cfg_ch->name_size);
					memcpy(name, msd_cfg_ch->name, tm);
					name[tm] = 0;
				}

				if (0 == http_proxy_size) {
					tm = snprintf(buf, sizeof(buf),
					    "#EXTINF:-1,%s\r\n%s\r\n\r\n",
					    name, name2);
				} else {
					if (0 == http_proxy_args_size) {
						tm = snprintf(buf, sizeof(buf),
						    "#EXTINF:-1,%s\r\n%s%s\r\n\r\n",
						    name, http_proxy, name2);
					} else {
						tm = snprintf(buf, sizeof(buf),
						    "#EXTINF:-1,%s\r\n%s%s?%s\r\n\r\n",
						    name, http_proxy, name2, http_proxy_args);
					}
				}
				write(out_fd, buf, tm);
			}
			write(out_fd, "\r\n\r\n", 4);
		}
	} else if (0 == buf_cmp(cmd, cmd_size, "astra_input_grep", 16)) {
		cnt = 0;
		for (i = 0; i < as_cfg_files_cnt; i ++) {
			for (j = 0; j < as_cfg_files[i].channels_cnt; j ++) {
				if (NULL == mem_find(0,
				    as_cfg_files[i].channels[j].input,
				    as_cfg_files[i].channels[j].input_size,
				    grep_arg, grep_arg_size))
					continue;
				fprintf(stdout, "%s\n",
				    as_cfg_files[i].channels[j].name);
				cnt ++;
			}
		}
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "astra_chk_missing_msd", 21)) {
		cnt = 0;
		for (i = 0; i < as_cfg_files_cnt; i ++) {
			for (j = 0; j < as_cfg_files[i].channels_cnt; j ++) {
				found = 0;
				/* look in msd sources. */
				if (NULL != msd_channel_find_by_addr(msd_cfg_files,
				    msd_cfg_files_cnt,
				    (as_cfg_files[i].channels[j].output + as_cfg_files[i].channels[j].output_udp_addr_offset),
				    (as_cfg_files[i].channels[j].output_size - as_cfg_files[i].channels[j].output_udp_addr_offset)))
					found = 1;
				if (0 == found) {
					fprintf(stdout, "Not found: %s (%s)\n",
					    as_cfg_files[i].channels[j].name,
					    (as_cfg_files[i].channels[j].output + as_cfg_files[i].channels[j].output_udp_addr_offset));
					cnt ++;
				}
			}
		}					
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "astra_and_msd", 13)) {
		cnt = 0;
		for (i = 0; i < as_cfg_files_cnt; i ++) {
			for (j = 0; j < as_cfg_files[i].channels_cnt; j ++) {
				found = 0;
				/* look in msd sources. */
				msd_cfg_ch = msd_channel_find_by_addr(msd_cfg_files,
				    msd_cfg_files_cnt,
				    (as_cfg_files[i].channels[j].output + as_cfg_files[i].channels[j].output_udp_addr_offset),
				    (as_cfg_files[i].channels[j].output_size - as_cfg_files[i].channels[j].output_udp_addr_offset));
				if (NULL != msd_cfg_ch) {
					found = 1;
					tm = min((sizeof(name) - 1), msd_cfg_ch->name_size);
					memcpy(name, msd_cfg_ch->name, tm);
					name[tm] = 0;

					fprintf(stdout, "%-32s %s\n",
					    as_cfg_files[i].channels[j].name, name);
					cnt ++;
				}
			}
		}					
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "msd_chk_missing_astra", 21)) {
		cnt = 0;
		for (i = 0; i < msd_cfg_files_cnt; i ++) {
			for (j = 0; j < msd_cfg_files[i].channels_cnt; j ++) {
				found = 0;
				for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt; k ++) {
					/* look in astra inputs. */
					if (NULL != as_cfg_find_by_addr(as_cfg_files,
					    as_cfg_files_cnt,
					    msd_cfg_files[i].channels[j].srcs[k].addr,
					    msd_cfg_files[i].channels[j].srcs[k].addr_size)) {
						found = 1;
						break;
					}
					if (0 != verbose) {
						tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
						memcpy(name, msd_cfg_files[i].channels[j].name, tm);
						name[tm] = 0;

						tm = min((sizeof(name2) - 1), msd_cfg_files[i].channels[j].srcs[k].addr_size);
						memcpy(name2, msd_cfg_files[i].channels[j].srcs[k].addr, tm);
						name2[tm] = 0;

						fprintf(stdout,"Not found: %s (%s)\n",
						    name, name2);
						cnt ++;
					}
				}
				if (0 == found && 0 == verbose) {
					tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
					memcpy(name, msd_cfg_files[i].channels[j].name, tm);
					name[tm] = 0;

					fprintf(stdout, "Not found: %s\n", name);
					cnt ++;
				}
			}
		}
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "msd_chk_missing_m3u", 19)) {
		cnt = 0;
		for (i = 0; i < msd_cfg_files_cnt; i ++) {
			for (j = 0; j < msd_cfg_files[i].channels_cnt; j ++) {
				found = 0;
				for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt; k ++) {
					/* look in m3u. */
					if (NULL != pl_m3u_find_by_addr(m3u_files, m3u_files_cnt,
					    msd_cfg_files[i].channels[j].srcs[k].addr,
					    msd_cfg_files[i].channels[j].srcs[k].addr_size)) {
						found = 1;
						break;
					}
					if (0 != verbose) {
						tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
						memcpy(name, msd_cfg_files[i].channels[j].name, tm);
						name[tm] = 0;

						tm = min((sizeof(name2) - 1), msd_cfg_files[i].channels[j].srcs[k].addr_size);
						memcpy(name2, msd_cfg_files[i].channels[j].srcs[k].addr, tm);
						name2[tm] = 0;

						fprintf(stdout,"Not found: %s (%s)\n",
						    name, name2);
						cnt ++;
					}
				}
				if (0 == found && 0 == verbose) {
					tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
					memcpy(name, msd_cfg_files[i].channels[j].name, tm);
					name[tm] = 0;

					fprintf(stdout, "Not found: %s\n", name);
					cnt ++;
				}
			}
		}
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "msd_chk_missing_astra_m3u", 25)) {
		cnt = 0;
		for (i = 0; i < msd_cfg_files_cnt; i ++) {
			for (j = 0; j < msd_cfg_files[i].channels_cnt; j ++) {
				found = 0;
				for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt; k ++) {
					/* look in astra inputs. */
					if (NULL != as_cfg_find_by_addr(as_cfg_files,
					    as_cfg_files_cnt,
					    msd_cfg_files[i].channels[j].srcs[k].addr,
					    msd_cfg_files[i].channels[j].srcs[k].addr_size)) {
						found = 1;
						break;
					}
					/* look in m3u. */
					if (NULL != pl_m3u_find_by_addr(m3u_files, m3u_files_cnt,
					    msd_cfg_files[i].channels[j].srcs[k].addr,
					    msd_cfg_files[i].channels[j].srcs[k].addr_size)) {
						found = 1;
						break;
					}
					if (0 != verbose) {
						tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
						memcpy(name, msd_cfg_files[i].channels[j].name, tm);
						name[tm] = 0;

						tm = min((sizeof(name2) - 1), msd_cfg_files[i].channels[j].srcs[k].addr_size);
						memcpy(name2, msd_cfg_files[i].channels[j].srcs[k].addr, tm);
						name2[tm] = 0;

						fprintf(stdout,"Not found: %s (%s)\n",
						    name, name2);
						cnt ++;
					}
				}
				if (0 == found && 0 == verbose) {
					tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
					memcpy(name, msd_cfg_files[i].channels[j].name, tm);
					name[tm] = 0;

					fprintf(stdout, "Not found: %s\n", name);
					cnt ++;
				}
			}
		}
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "msd_and_astra", 13)) {
		cnt = 0;
		for (i = 0; i < msd_cfg_files_cnt; i ++) {
			for (j = 0; j < msd_cfg_files[i].channels_cnt; j ++) {
				found = 0;
				for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt && 0 == found; k ++) {
					/* look in astra inputs. */
					as_cfg_ch = as_cfg_find_by_addr(as_cfg_files,
					    as_cfg_files_cnt,
					    msd_cfg_files[i].channels[j].srcs[k].addr,
					    msd_cfg_files[i].channels[j].srcs[k].addr_size);
					if (NULL != as_cfg_ch) {
						found = 1;
						tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
						memcpy(name, msd_cfg_files[i].channels[j].name, tm);
						name[tm] = 0;

						fprintf(stdout, "%-32s %s\n", name, as_cfg_ch->name);
						cnt ++;
					}					
				}
			}
		}
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "msd_and_m3u", 11)) {
		cnt = 0;
		for (i = 0; i < msd_cfg_files_cnt; i ++) {
			for (j = 0; j < msd_cfg_files[i].channels_cnt; j ++) {
				found = 0;
				for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt && 0 == found; k ++) {
					/* look in m3u. */
					m3u_ch = pl_m3u_find_by_addr(m3u_files, m3u_files_cnt,
					    msd_cfg_files[i].channels[j].srcs[k].addr,
					    msd_cfg_files[i].channels[j].srcs[k].addr_size);
					if (NULL != m3u_ch) {
						found = 1;
						tm = min((sizeof(name) - 1), msd_cfg_files[i].channels[j].name_size);
						memcpy(name, msd_cfg_files[i].channels[j].name, tm);
						name[tm] = 0;

						tm = min((sizeof(name2) - 1), m3u_ch->name_size);
						memcpy(name2, m3u_ch->name, tm);
						name2[tm] = 0;

						fprintf(stdout, "%-32s %s\n", name, name2);
						cnt ++;
					}					
				}
			}
		}
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "m3u_chk_missing_msd", 19)) {
		cnt = 0;
		for (i = 0; i < m3u_files_cnt; i ++) {
			for (j = 0; j < m3u_files[i].channels_cnt; j ++) {
				found = 0;
				/* look in msd sources. */
				if (NULL != msd_channel_find_by_addr(msd_cfg_files,
				    msd_cfg_files_cnt,
				    (m3u_files[i].channels[j].url + m3u_files[i].channels[j].url_udp_addr_offset),
				    (m3u_files[i].channels[j].url_size - m3u_files[i].channels[j].url_udp_addr_offset)))
					found = 1;
				if (0 == found) {
					tm = min((sizeof(name) - 1), m3u_files[i].channels[j].name_size);
					memcpy(name, m3u_files[i].channels[j].name, tm);
					name[tm] = 0;

					tm = min((sizeof(name2) - 1), (m3u_files[i].channels[j].url_size - m3u_files[i].channels[j].url_udp_addr_offset));
					memcpy(name2, (m3u_files[i].channels[j].url + m3u_files[i].channels[j].url_udp_addr_offset), tm);
					name2[tm] = 0;

					fprintf(stdout, "Not found: %s (%s)\n",
					    name, name2);
					cnt ++;
				}
			}
		}					
		fprintf(stdout, "Total: %zu\n", cnt);
	} else if (0 == buf_cmp(cmd, cmd_size, "m3u_and_msd", 11)) {
		cnt = 0;
		for (i = 0; i < m3u_files_cnt; i ++) {
			for (j = 0; j < m3u_files[i].channels_cnt; j ++) {
				found = 0;
				/* look in msd sources. */
				msd_cfg_ch = msd_channel_find_by_addr(msd_cfg_files,
				    msd_cfg_files_cnt,
				    (m3u_files[i].channels[j].url + m3u_files[i].channels[j].url_udp_addr_offset),
				    (m3u_files[i].channels[j].url_size - m3u_files[i].channels[j].url_udp_addr_offset));
				if (NULL != msd_cfg_ch) {
					found = 1;
					tm = min((sizeof(name) - 1), msd_cfg_ch->name_size);
					memcpy(name, msd_cfg_ch->name, tm);
					name[tm] = 0;

					tm = min((sizeof(name2) - 1), m3u_files[i].channels[j].name_size);
					memcpy(name2, m3u_files[i].channels[j].name, tm);
					name2[tm] = 0;

					fprintf(stdout, "%-32s %s\n", name2, name);
					cnt ++;
				}
			}
		}					
		fprintf(stdout, "Total: %zu\n", cnt);
	} else {
		fprintf(stderr, "Uncnown command!\n");
		fprintf(stderr, "manual: http://www.netlab.linkpc.net/wiki/ru:software:msd:cfg_tool\n");
		fprintf(stderr, "manual: http://netlab.dhis.org/wiki/ru:software:msd:cfg_tool\n");
		fprintf(stderr, "manual: http://net-lab.slyip.net/wiki/ru:software:msd:cfg_tool\n");
	}

	/* No mem free() here. */

	close(out_fd);

	return (0);
}


int
load_file(int fd, void **buf, size_t *buf_size) {
	uint8_t *data = NULL;
	ssize_t size;

	if (-1 == fd || NULL == buf)
		return (EINVAL);
	/* Get file size. */
	size = lseek(fd, 0, SEEK_END);
	if (-1 == size) {
		LOG_ERR(errno, "lseek()");
		return (errno);
	}
	lseek(fd, 0, SEEK_SET);
	/* Allocate buf for file content. */
	data = malloc((size + 4));
	if (NULL == data) {
		LOG_ERR(errno, "malloc()");
		return (errno);
	}
	/* Read file content. */
	size = read(fd, data, size);
	if (-1 == size) {
		free(data);
		LOG_ERR(errno, "read()");
		return (errno);
	}
	data[size] = 0;

	(*buf) = data;
	if (NULL != buf_size)
		(*buf_size) = size;

	return (0);
}

int
skeep_spwsp(uint8_t *buf, size_t buf_size,
    uint8_t **buf_ret, size_t *buf_size_ret) {
	uint8_t *buf_max;

	if (NULL == buf && 0 != buf_size)
		return (EINVAL);
	buf_max = (buf + buf_size);
	/* Skeep head spaces. */
	for (; 33 > (*buf) && buf < buf_max; buf ++)
		;
	if (NULL != buf_ret)
		(*buf_ret) = buf;
	if (NULL != buf_size_ret)
		(*buf_size_ret) = (buf_max - buf);
	return (0);
}
int
skeep_spwsp2(uint8_t *buf, size_t buf_size, uint8_t **buf_ret, size_t *buf_size_ret) {
	uint8_t *buf_max;

	if (NULL == buf && 0 != buf_size)
		return (EINVAL);
	buf_max = (buf + buf_size - 1);
	if (NULL != buf_ret) {
		/* Skeep head spaces. */
		for (; 33 > (*buf) && buf <= buf_max; buf ++)
			;
		(*buf_ret) = buf;
	}
	if (NULL != buf_size_ret) {
		/* Skeep tail spaces. */
		for (; 33 > (*buf_max) && buf <= buf_max; buf_max --)
			;
		(*buf_size_ret) = ((buf_max + 1) - buf);
	}
	return (0);
}

int
buf_get_next_line(uint8_t *buf, size_t buf_size,
    uint8_t *line, size_t line_size,
    uint8_t **next_line, size_t *next_line_size) {
	uint8_t *ptr, *end, *buf_end = (buf + buf_size);

	if (NULL == buf || 0 == buf_size)
		return (EINVAL);
	if (NULL != line) {
		ptr = (line + line_size);
		if (ptr < buf)
			ptr = buf;
		if (0x0d == ptr[0]) /*CR*/
			ptr ++;
		if (0x0a == ptr[0]) /*LF*/
			ptr ++;
		if (ptr > buf_end)
			ptr = buf_end;
	} else {
		ptr = buf;
	}

	end = mem_find_byte((ptr - buf), buf, buf_size, 0x0a/*LF*/);
	if (NULL == end)
		end = buf_end;
	else if (end > ptr && 0x0d == end[-1]) /*CR*/
		end --;
	if (buf_end == ptr) /* EOF */
		return (-1);

	(*next_line) = ptr;
	(*next_line_size) = (end - ptr);

	return (0);
}
uint8_t *
buf_skeep_non_print(uint8_t *buf, size_t buf_size, uint8_t *ptr) {
	uint8_t *buf_end = (buf + buf_size);

	if (NULL == ptr || buf > ptr)
		ptr = buf;
	for (; ptr < buf_end && (*ptr) <= ' '; ptr ++)
		;
	return (ptr);
}



as_cfg_ch_p
as_cfg_find_by_addr(as_cfg_file_p as_cfg_files, size_t as_cfg_files_cnt,
    uint8_t *addr, size_t addr_size) {
	size_t i, j;

	for (i = 0; i < as_cfg_files_cnt; i ++) {
		for (j = 0; j < as_cfg_files[i].channels_cnt; j ++) {
			if (0 == buf_cmp(addr, addr_size,
			    (as_cfg_files[i].channels[j].output + as_cfg_files[i].channels[j].output_udp_addr_offset),
			    (as_cfg_files[i].channels[j].output_size - as_cfg_files[i].channels[j].output_udp_addr_offset)))
				return (&as_cfg_files[i].channels[j]);
		}
	}

	return (NULL);
}

int
as_cfg_load_dir(char *dir_name, as_cfg_file_p as_cfg_files, size_t *as_cfg_files_cnt) {
	int root_fd, fd, i, dir_cnt;
	struct dirent *de, **namelist;

	root_fd = open(dir_name, O_RDONLY);
	if (-1 == root_fd) {
		LOG_ERR(errno, "open()");
		return (errno);
	}
	dir_cnt = scandir(dir_name, &namelist, 0, alphasort);
	if (dir_cnt < 0) {
		LOG_ERR(errno, "scandir()");
		return (errno);
	}

	for (i = 0; i < dir_cnt; i ++) {
		de = namelist[i];
		if (0 == buf_cmp(de->d_name, _D_EXACT_NAMLEN(de), ".", 1) ||
		    0 == buf_cmp(de->d_name, _D_EXACT_NAMLEN(de), "..", 2))
			continue;
		switch (de->d_type) {
		case DT_DIR:
			break;
		default:
			if (MAX_CFG_FILES <= (*as_cfg_files_cnt))
				break;
			fd = openat(root_fd, de->d_name, O_RDONLY);
			if (-1 == fd) {
				LOG_ERR_FMT(errno, "openat(%s%s)",
				    dir_name, de->d_name);
				break;
			}
			if (0 != as_cfg_file_parse(fd, &as_cfg_files[(*as_cfg_files_cnt)]))
				break;
			(*as_cfg_files_cnt) ++;
			break;
		}
		free(de);
	}
	free(namelist);
	close(root_fd);

	return (0);
}
int
as_cfg_load_file(char *file_name, as_cfg_file_p cfg_file) {
	int fd;

	fd = open(file_name, O_RDONLY);
	if (-1 == fd) {
		LOG_ERR_FMT(errno, "open(%s)", file_name);
		return (errno);
	}
	return (as_cfg_file_parse(fd, cfg_file));
}

int
as_cfg_file_parse(int fd, as_cfg_file_p cfg_file) {
	int error = 0, ch_single_val;
	uint8_t *fcont = NULL, *cur_pos, *ch_start, *ch_end;
	as_cfg_ch_p chann, channs = NULL, tmchanns;
	size_t fsize = 0, channs_cnt = 0, channs_alloc = 0;
	uint8_t *ch_cur, *ptm, *ch_val_name, *ch_val_data;
	size_t ch_size, ch_val_name_size, ch_val_data_size;


	error = load_file(fd, (void**)&fcont, &fsize);
	close(fd);
	if (0 != error) {
		LOG_ERR(error, "load_file()");
		goto err_out;
	}
	
	/* enum channels. */
	cur_pos = fcont;
	for (;;) {
		ch_start = mem_find((cur_pos - fcont), fcont, fsize,
		    "make_channel({", 14);
		if (NULL == ch_start)
			break;
		ch_start += 14;
		ch_end = mem_find((ch_start - fcont), fcont, fsize,
		    "})", 2);
		if (NULL == ch_end)
			break;
		ch_size = (ch_end - ch_start);
		cur_pos = (ch_end + 2);
		
		if (NULL == channs || channs_cnt == channs_alloc) {
			channs_alloc += 16;
			tmchanns = realloc(channs,
			    (sizeof(as_cfg_ch_t) * channs_alloc));
			if (NULL == tmchanns) {
				error = errno;
				LOG_ERR(error, "realloc()");
				goto err_out;
			}
			channs = tmchanns;
		}
		chann = &channs[channs_cnt];
		chann->enable = 1;

		/* Parce channel. */
		ch_cur = ch_start;
		for (; NULL != ch_cur;) {
			/* get val name. */
			ptm = mem_find_byte((ch_cur - ch_start),
			    ch_start, ch_size, '=');
			if (NULL == ptm)
				break;
			skeep_spwsp((ptm + 1), (ch_size - ((ptm + 1) - ch_start)),
			    &ch_val_data, NULL); /* Save value data start. */
			ptm --;
			skeep_spwsp2(ch_cur, (ptm - ch_cur), &ch_val_name, &ch_val_name_size);
			ch_val_name[ch_val_name_size] = 0;
			
			/* get val data. */
			if ('{' == ch_val_data[0]) {
				ptm = mem_find_byte((ch_val_data - ch_start),
				    ch_start, ch_size, '}');
				if (NULL == ptm)
					break; /* Bad format. */
				ch_cur = mem_find_byte((ptm - ch_start),
				    ch_start, ch_size, ',');
				if (NULL != ch_cur)
					ch_cur ++;
				ch_val_data ++; /* remove braces. */
				if (NULL != mem_find_byte(0, ch_val_data,
				    (ptm - ch_val_data), ',')) {
					ch_single_val = 0;
				} else {
					ch_single_val = 1;
				}
			} else {
				ptm = mem_find_byte((ch_val_data - ch_start),
				     ch_start, ch_size, ',');
				if (NULL == ptm) { /* Last value. */
					ptm = (ch_start + ch_size);
					ch_cur = NULL;
				} else {
					ch_cur = (ptm + 1);
				}
				ch_single_val = 1;
			}
			ch_val_data_size = (ptm - ch_val_data);
			/* Remove sp/tab/" */
			skeep_spwsp2(ch_val_data, ch_val_data_size,
			    &ch_val_data, &ch_val_data_size);
			if (ch_single_val) {
				if ('"' == ch_val_data[0]) {
					ch_val_data ++;
					ch_val_data_size -= 2;
				}
			}
			ch_val_data[ch_val_data_size] = 0;

			/* Store to struct by name. */
			if (0 == buf_cmp(ch_val_name, ch_val_name_size,
			    "name", 4)) {
				chann->name = ch_val_data;
				chann->name_size = ch_val_data_size;
			} else if (0 == buf_cmp(ch_val_name, ch_val_name_size,
			    "input", 5)) {
				chann->input = ch_val_data;
				chann->input_size = ch_val_data_size;
			} else if (0 == buf_cmp(ch_val_name, ch_val_name_size,
			    "output", 6)) {
				chann->output = ch_val_data;
				chann->output_size = ch_val_data_size;
				if (8 < chann->output_size &&
				    (0 == buf_cmp(chann->output, 6, "udp://", 6) ||
				    0 == buf_cmp(chann->output, 6, "rtp://", 6))) {
					ptm = mem_find_byte(6, chann->output,
					    chann->output_size, '@');
					if (NULL != ptm) {
						ptm ++;
					} else {
						ptm = (chann->output + 6);
					}
					chann->output_udp_addr_offset = (ptm - chann->output);
				} else {
					chann->output_udp_addr_offset = 0;
				}
			} else if (0 == buf_cmp(ch_val_name, ch_val_name_size,
			    "enable", 6) && 0 != ch_val_data_size) {
				switch (ch_val_data[0]) {
				case 'f':
				case 'F':
				case 'n':
				case 'N':
				case '0':
					chann->enable = 0;
					break;
				}
			}
		}
		
		if (NULL != chann->name)
			channs_cnt ++;
	}
	
err_out:
	if (0 != error) {
		free(fcont);
		fcont = NULL;
		
		free(channs);
		channs = NULL;
		channs_cnt = 0;
	}

	cfg_file->channels = channs;
	cfg_file->channels_cnt = channs_cnt;
	cfg_file->buf = fcont;
	
	return (error);
}



msd_cfg_ch_p
msd_channel_find_by_addr(msd_cfg_file_p msd_cfg_files, size_t msd_cfg_files_cnt,
    uint8_t *addr, size_t addr_size) {
	size_t i, j, k;

	/* look in msd sources. */
	for (i = 0; i < msd_cfg_files_cnt; i ++) {
		for (j = 0; j < msd_cfg_files[i].channels_cnt; j ++) {
			for (k = 0; k < msd_cfg_files[i].channels[j].srcs_cnt; k ++) {
				if (0 == buf_cmp(addr, addr_size,
				    msd_cfg_files[i].channels[j].srcs[k].addr,
				    msd_cfg_files[i].channels[j].srcs[k].addr_size))
					return (&msd_cfg_files[i].channels[j]);
			}
		}
	}
	return (NULL);
}

int
msd_channel_load(uint8_t *data, size_t data_size, msd_cfg_ch_p channel) {
	uint8_t *src_data, *cur_pos;
	size_t src_data_size;

	/* name */
	if (0 != xml_get_val_args(data, data_size, NULL, NULL, NULL,
	    &channel->name, &channel->name_size, (const uint8_t*)"name", NULL))
		return (EINVAL);

	/* hubProfileName - load hub params from template. */
	xml_get_val_args(data, data_size, NULL, NULL, NULL,
	    &channel->profile_name, &channel->profile_name_size,
	    (const uint8_t*)"hubProfileName", NULL);

	/* Stream hub sources. */
	cur_pos = NULL;
	channel->srcs_cnt = 0;
	while (0 == xml_get_val_args(data, data_size, &cur_pos, NULL, NULL,
	    &src_data, &src_data_size,
	    (const uint8_t*)"sourceList", "source", NULL) &&
	    MAX_MSD_SRCS > channel->srcs_cnt) {
		/* type */
		if (0 != xml_get_val_args(src_data, src_data_size,
		    NULL, NULL, NULL,
		    &channel->srcs[channel->srcs_cnt].type,
		    &channel->srcs[channel->srcs_cnt].type_size,
		    (const uint8_t*)"type", NULL))
			continue;

		/* sourceProfileName. */
		xml_get_val_args(src_data, src_data_size,
		    NULL, NULL, NULL,
		    &channel->srcs[channel->srcs_cnt].profile_name,
		    &channel->srcs[channel->srcs_cnt].profile_name_size,
		    (const uint8_t*)"sourceProfileName", NULL);

		/* addr. */
		channel->srcs[channel->srcs_cnt].addr = NULL;
		channel->srcs[channel->srcs_cnt].addr_size = 0;
		channel->srcs[channel->srcs_cnt].addr_udp = 0;
		if (NULL != mem_find(0, channel->srcs[channel->srcs_cnt].type,
		    channel->srcs[channel->srcs_cnt].type_size,
		    "udp", 3)) {
			xml_get_val_args(src_data, src_data_size,
			    NULL, NULL, NULL,
			    &channel->srcs[channel->srcs_cnt].addr,
			    &channel->srcs[channel->srcs_cnt].addr_size,
			    (const uint8_t*)"udp", "address", NULL);
			channel->srcs[channel->srcs_cnt].addr_udp = 1;
		} else if (NULL != mem_find(0, channel->srcs[channel->srcs_cnt].type,
		    channel->srcs[channel->srcs_cnt].type_size,
		    "tcp", 3)) {
			/* XXX: need read 2 parts. */
			xml_get_val_args(src_data, src_data_size,
			    NULL, NULL, NULL,
			    &channel->srcs[channel->srcs_cnt].addr,
			    &channel->srcs[channel->srcs_cnt].addr_size,
			    (const uint8_t*)"http", "urlPath", NULL);
		}
		channel->srcs_cnt ++;
	} /* End channel sources load. */

	return (0);
}
int
msd_cfg_load_channels(uint8_t *buf, size_t buf_size, msd_cfg_file_p cfg_file) {
	uint8_t *data, *cur_pos;
	size_t data_size, channs_alloc = 0;
	msd_cfg_ch_p tmchanns;

	/* Load channels. */
	cur_pos = NULL;
	cfg_file->channels = NULL;
	cfg_file->channels_cnt = 0;

	while (0 == xml_get_val_args(buf, buf_size, &cur_pos,
	    NULL, NULL, &data, &data_size,
	    (const uint8_t*)"channel", NULL)) {
		if (NULL == cfg_file->channels ||
		    cfg_file->channels_cnt == channs_alloc) {
			channs_alloc += 16;
			tmchanns = realloc(cfg_file->channels,
			    (sizeof(msd_cfg_ch_t) * channs_alloc));
			if (NULL == tmchanns) {
				free(cfg_file->channels);
				LOG_ERR(errno, "realloc()");
				return (errno);
			}
			cfg_file->channels = tmchanns;
		}
		if (0 == msd_channel_load(data, data_size,
		    &cfg_file->channels[cfg_file->channels_cnt]))
			cfg_file->channels_cnt ++;
	}

	return (0);
}

int
msd_cfg_load_file(char *file_name, msd_cfg_file_p msd_cfg_files, size_t *msd_cfg_files_cnt) {
	int fd, error;
	uint8_t *cur_pos, *fdata, *data;
	char strbuf[4096];
	size_t fdata_size, data_size;

	if (MAX_CFG_FILES <= (*msd_cfg_files_cnt))
		return (-1);

	fd = open(file_name, O_RDONLY);
	if (-1 == fd) {
		LOG_ERR(errno, "open()");
		return (errno);
	}
	error = load_file(fd, (void**)&fdata, &fdata_size);
	close(fd);
	if (0 != error) {
		LOG_ERR(error, "load_file()");
		return (error);
	}

	msd_cfg_files[(*msd_cfg_files_cnt)].channels = NULL;
	msd_cfg_files[(*msd_cfg_files_cnt)].channels_cnt = 0;
	msd_cfg_files[(*msd_cfg_files_cnt)].buf = fdata;

	/* Load channels. */
	/* From main config. */
	if (0 == xml_get_val_args(fdata, fdata_size, NULL,
	    NULL, NULL, &data, &data_size,
	    (const uint8_t*)"msd", "channelList", NULL)) {
		if (0 == msd_cfg_load_channels(data, data_size,
		    &msd_cfg_files[(*msd_cfg_files_cnt)]))
			(*msd_cfg_files_cnt) ++;
	} else { /* Channels only file. */
		if (0 == msd_cfg_load_channels(fdata, fdata_size,
		    &msd_cfg_files[(*msd_cfg_files_cnt)])) {
			(*msd_cfg_files_cnt) ++;
		} else {
			free(fdata);
		}
		return (0);
	}
	/* From included configs. */
	cur_pos = NULL;
	while (0 == xml_get_val_args(fdata, fdata_size, &cur_pos,
	    NULL, NULL, &data, &data_size,
	    (const uint8_t*)"msd", "channelList", "includeFile", NULL) &&
	    MAX_CFG_FILES > (*msd_cfg_files_cnt)) {
		if ((sizeof(strbuf) - 1) < data_size)
			continue;
		memcpy(strbuf, data, data_size);
		strbuf[data_size] = 0;
		fd = open(strbuf, O_RDONLY);
		if (-1 == fd) {
			LOG_ERR(errno, "open()");
			continue;
		}
		error = load_file(fd, (void**)&data, &data_size);
		close(fd);
		if (0 != error) {
			LOG_ERR(error, "load_file()");
			continue;
		}
		msd_cfg_files[(*msd_cfg_files_cnt)].channels = NULL;
		msd_cfg_files[(*msd_cfg_files_cnt)].channels_cnt = 0;
		msd_cfg_files[(*msd_cfg_files_cnt)].buf = data;
		if (0 == msd_cfg_load_channels(data, data_size,
		    &msd_cfg_files[(*msd_cfg_files_cnt)]))
			(*msd_cfg_files_cnt) ++;
	}

	return (0);
}



m3u_ch_p
pl_m3u_find_by_addr(m3u_file_p m3u_files, size_t m3u_files_cnt,
    uint8_t *addr, size_t addr_size) {
	size_t i, j;

	/* look in m3u. */
	for (i = 0; i < m3u_files_cnt; i ++) {
		for (j = 0; j < m3u_files[i].channels_cnt; j ++) {
			if (0 == buf_cmp(addr, addr_size,
			    (m3u_files[i].channels[j].url + m3u_files[i].channels[j].url_udp_addr_offset),
			    (m3u_files[i].channels[j].url_size - m3u_files[i].channels[j].url_udp_addr_offset)))
				return (&m3u_files[i].channels[j]);
		}
	}
	return (NULL);
}

int
pl_m3u_buf_load(char *file_name, m3u_file_p m3u_file) {
	int error, fd;
	uint8_t *buf, *hdr_line, *url_line, *val, *ptr;
	size_t buf_size, hdr_line_size, hdr_line_data_size, url_line_size, val_size;
	size_t channs_alloc = 0;
	m3u_ch_p tmchanns;


	fd = open(file_name, O_RDONLY);
	if (-1 == fd) {
		LOG_ERR(errno, "open()");
		return (errno);
	}
	error = load_file(fd, (void**)&buf, &buf_size);
	close(fd);
	if (0 != error) {
		LOG_ERR(error, "load_file()");
		return (error);
	}

	m3u_file->channels = NULL;
	m3u_file->channels_cnt = 0;
	m3u_file->buf = buf;


	/* First line. */
	hdr_line = NULL;
	hdr_line_size = 0;
	if (0 != buf_get_next_line(buf, buf_size, hdr_line, hdr_line_size, &hdr_line, &hdr_line_size) ||
	    7 > hdr_line_size ||
	    NULL == (ptr = mem_find(0, buf, buf_size, "#EXTM3U", 7))) {
		error = -1;
		goto err_out;
	}

	/* First line looks OK. */
	/* Skeep BOM. */
	hdr_line_size -= (ptr - hdr_line);
	hdr_line = ptr;
	
	/* Handle items. */
	while (0 == buf_get_next_line(buf, buf_size, hdr_line, hdr_line_size, &hdr_line, &hdr_line_size)) {
		if (8 >= hdr_line_size || 0 != memcmp(hdr_line, "#EXTINF:", 8))
			continue; /* Skeep unknown. */
		while (0 == buf_get_next_line(buf, buf_size, hdr_line, hdr_line_size, &url_line, &url_line_size)) {
			if (0 == hdr_line_size)
				continue; /* Skeep empty line. */
			if (8 >= url_line_size || 0 != memcmp(url_line, "#EXTINF:", 8))
				break;
			/* Skeep item without url. */
			hdr_line = url_line;
			hdr_line_size = url_line_size;
		}
		hdr_line += 8;
		hdr_line_size -= 8;
		
		/* Get item name. */
		val = mem_find_byte_rev(0, hdr_line, hdr_line_size, ',');
		if (NULL == val)
			continue; /* Invalid line format. */
		hdr_line_data_size = (val - hdr_line);
		val ++;
		val = buf_skeep_non_print(hdr_line, hdr_line_size, val);
		val_size = ((hdr_line + hdr_line_size) - val);
		
		/* Add Channel. */
		if (NULL == m3u_file->channels ||
		    m3u_file->channels_cnt == channs_alloc) {
			channs_alloc += 16;
			tmchanns = realloc(m3u_file->channels,
			    (sizeof(msd_cfg_ch_t) * channs_alloc));
			if (NULL == tmchanns) {
				free(m3u_file->channels);
				LOG_ERR(errno, "realloc()");
				return (errno);
			}
			m3u_file->channels = tmchanns;
		}
		/* strip spaces. */
		skeep_spwsp2(url_line, url_line_size, &url_line, &url_line_size);

		m3u_file->channels[m3u_file->channels_cnt].name = val;
		m3u_file->channels[m3u_file->channels_cnt].name_size = val_size;
		m3u_file->channels[m3u_file->channels_cnt].url = url_line;
		m3u_file->channels[m3u_file->channels_cnt].url_size = url_line_size;
		if (8 < url_line_size &&
		    (0 == buf_cmp(url_line, 6, "udp://", 6) ||
		    0 == buf_cmp(url_line, 6, "rtp://", 6))) {
			ptr = mem_find_byte(6, url_line, url_line_size, '@');
			if (NULL != ptr) {
				ptr ++;
			} else {
				ptr = (url_line + 6);
			}
			m3u_file->channels[m3u_file->channels_cnt].url_udp_addr_offset = (ptr - url_line);
		} else {
			m3u_file->channels[m3u_file->channels_cnt].url_udp_addr_offset = 0;
		}
		m3u_file->channels_cnt ++;
	}

	error = 0;

err_out:
	return (error);
}
